/**************************************************************************/
/*                                                                        */
/*  This file is part of Frama-Clang                                      */
/*                                                                        */
/*  Copyright (C) 2012-2017                                               */
/*    CEA (Commissariat à l'énergie atomique et aux énergies              */
/*         alternatives)                                                  */
/*                                                                        */
/*  you can redistribute it and/or modify it under the terms of the GNU   */
/*  Lesser General Public License as published by the Free Software       */
/*  Foundation, version 2.1.                                              */
/*                                                                        */
/*  It is distributed in the hope that it will be useful,                 */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*  GNU Lesser General Public License for more details.                   */
/*                                                                        */
/*  See the GNU Lesser General Public License version 2.1                 */
/*  for more details (enclosed in the file LICENSE).                      */
/*                                                                        */
/**************************************************************************/

//
// Description:
//  Implementation of the ACSL++ lexer.
//

#include <sstream>
#include <climits>
#include "ACSLLexer.h"
#include "intermediate_format.h"

namespace Acsl {

using namespace DLexer;

std::string Lexer::utf8_convert(unsigned c) {
  if (c<0x80) { return std::string(1,c); }
  unsigned length = 0;
  unsigned i = c>>1;
  while (i!=0) { length++; i>>=5; }
  char first = (((1 << length) -1) << (8-length)) | (c>>(6*(length-1)));
  std::string res(length,first);
  for (i=1; i<length; i++) res[i] = (((c>>((length - 1 -i)*6)) & 0x3f) | 0x80);
  return res;
}

Lexer::Error::Error(location position, const std::string& msg)
  : filePos(position->filename1), linePos(position->linenum1),
    columnPos(position->charnum1), message(msg) {}

Lexer::ReadResult
Lexer::readEndComment(const std::string& buffer, size_t& position, location loc,
    bool doesForce) {
  size_t pos;
  if (_context == '*')
    pos = buffer.find("*/", position);
  else
    pos = buffer.find('\n', position);
  unsigned lineEnd = loc->linenum2;
  unsigned charEnd = loc->charnum2;
  if (_context == '*') {
    unsigned previousCharEnd = charEnd;
    size_t subpos = buffer.find('\n', position);
    if (subpos != std::string::npos
        && (pos == std::string::npos || subpos < pos)) {
      ++lineEnd;
      previousCharEnd = (subpos-position)+charEnd;
      charEnd = 1;
      size_t subsubpos;
      bool doesContinue;
      do {
        subsubpos = buffer.find('\n', subpos);
        doesContinue = (subsubpos != std::string::npos
            && (pos == std::string::npos || subsubpos < pos));
        if (doesContinue) {
          ++lineEnd;
          previousCharEnd = (subsubpos-subpos)+1;
          subpos = subsubpos;
        };
      } while (doesContinue);
      charEnd = (pos != std::string::npos) ? (pos-subpos)+2
                                           : (buffer.length()-subpos-1);
      if (charEnd == 0) {
        charEnd = previousCharEnd;
        if (lineEnd > loc->linenum2)
          --lineEnd;
      };
    }
    else
      charEnd += (pos != std::string::npos)
          ? (pos-position)+1 : (buffer.length()-position-1);
  }
  else
    charEnd += (pos != std::string::npos)
        ? (pos-position) : (buffer.length()-position-1);


  if (pos != std::string::npos) {
    _currentToken.readFrom(buffer, position, pos-position);
    position = pos;
    if (_context == '*')
      position += 2;
    else {
      position++;
      _hasNewlineToken = true;
    };
    CommentToken* result = new CommentToken;
    _token = Token(result);
    _currentToken >> result->content();
    _secondaryState = NULL;
    loc->linenum2 = lineEnd;
    loc->charnum2 = charEnd;
    return RRHasToken;
  };

  if (!doesForce) {
    assert(buffer.size()-position > 1);
    _currentToken.readFrom(buffer, position, buffer.size()-position-1);
    position = buffer.size()-1;
    loc->linenum2 = lineEnd;
    loc->charnum2 = charEnd;
    return RRNeedChars;
  };
  
  loc->linenum2 = lineEnd;
  loc->charnum2 = charEnd+1;
  if (buffer.length() > 0 && buffer[buffer.length()-1] == '\n') {
    ++loc->linenum2;
    loc->charnum2 = 1;
  };

  if (_context == '*') {
    _errorList.push_back(Error(loc, "comment not terminated"));
    return RRFinished;
  };
    
  _currentToken.readFrom(buffer, position, buffer.size()-position);
  position = buffer.size();
  CommentToken* result = new CommentToken;
  _token = Token(result);
  _currentToken >> result->content();
  _secondaryState = NULL;
  loc->linenum2 = lineEnd;
  loc->charnum2 = charEnd;
  return RRHasToken;
}

Lexer::ReadResult
Lexer::readProtectedToken(const std::string& buffer, size_t& position,
    location loc, bool doesForce) {
  char ch = buffer[position];
  while (isalnum(ch) || ch == '_') {
    _currentToken << ch;
    ++position;
    ++loc->charnum2;
    ch = buffer[position];
  };
  if (ch != '\0' || doesForce) {
    std::string result;
    _currentToken >> result;
    KeywordToken::Map::const_iterator
      found = KeywordToken::_protectedKeywords.find(result);
    if (found != KeywordToken::_protectedKeywords.end()) {
      KeywordToken result;
      result.setType(found->second);
      _token = Token(result);
    }
    else {
      // treat it as a (built-in) identifier
      IdentifierToken* ident = new IdentifierToken;
      _token = Token(ident);
      ident->content() = "\\" + result;
    };
    _secondaryState = NULL;
    return RRHasToken;
  };
  return RRNeedChars;
}
  
Lexer::ReadResult
Lexer::readIdentifierToken(const std::string& buffer, size_t& position,
    location loc, bool doesForce) {
  char ch = buffer[position];
  while (isalnum(ch) || ch == '_') {
    _currentToken << ch;
    ++position;
    ++loc->charnum2;
    ch = buffer[position];
  };
  if (ch != '\0' || doesForce) {
    std::string ident;
    _currentToken >> ident;

    KeywordToken::Map::const_iterator found = 
      KeywordToken::_unprotectedKeywords.find(ident);
    if (found != KeywordToken::_unprotectedKeywords.end()) {
      KeywordToken result;
      result.setType(found->second);
      _token = Token(result);
    }
    else {
      IdentifierToken* result = new IdentifierToken;
      _token = Token(result);
      result->content() = ident;
    };
    _secondaryState = NULL;
    return RRHasToken;
  };
  return RRNeedChars;
}

Lexer::ReadResult
Lexer::readNumberToken(const std::string& buffer, size_t& position,
    location loc, bool doesForce) {
  while(buffer[position] != '\0' || doesForce) {
    char ch = buffer[position];
    switch (_litState) {
      case LeadZero: {
        if ( ch == 'x' || ch == 'X' ) {
          _digits = Hex;
          _litState = Digits;
          advanceChar(buffer,position,loc);
          continue;
        }
        if (ch == 'b' || ch == 'B' ) {
          _digits = Bin;
          _litState = Digits;
          advanceChar(buffer,position,loc);
          continue;
        }
        if (ch == '.') {
          if (buffer[position+1] == '.') {
            // special case: 0..xxx is a range of integers.
            // Just return the lexeme 0 as an int constant.
            IntegerLiteralToken* zero = new IntegerLiteralToken;
            zero->setType(IntegerLiteralToken::TOctal);
            _currentToken >> zero->value();
            _token = Token(zero);
            return RRHasToken;
          }
          _digits = Dec;
          _litState = FracPart;
          advanceChar(buffer,position,loc);
          continue;
        }
        _digits = Oct;
        _litState = Digits;
        continue;
      }
      case Digits: {
        while (isDigit(ch)) {
          advanceChar(buffer,position,loc);
          ch = buffer[position]; // = continue;
          if (ch == '\0' && !doesForce)
            return RRNeedChars;
        }
        // We only support decimal or hexadecimal floats
        if (ch == '.' && (_digits != Oct)) {
          if (buffer[position+1] == '.') {
            // special case: xxx..expr is a range of integers.
            // Just return the lexeme xxx as an int constant.
            IntegerLiteralToken* cst = new IntegerLiteralToken;
            setType(cst);
            _currentToken >> cst->value();
            _token = Token(cst);
            return RRHasToken;
          }
          _litState = FracPart;
          advanceChar(buffer,position,loc);
          continue;
        }
        if ((ch == 'e' || ch == 'E') && (_digits == Dec || _digits == Bin)) {
          _litState = Exponent;
          advanceChar(buffer,position,loc);
          continue;
        }
        if ((ch == 'p' || ch == 'P') && _digits == Hex) {
          _litState = Exponent;
          advanceChar(buffer,position,loc);
          continue;
        }
        // We are seeing an integer. It remains to check for a suffix.
        IntegerLiteralToken* intToken = new IntegerLiteralToken;
        setType(intToken);
        _token = Token(intToken);
        _litState = ISuf;
        continue;
      }
      case FracPart: {
        while (isDigit(ch)) {
          advanceChar(buffer,position,loc);
          ch = buffer[position]; // = continue;
          if (ch == '\0' && !doesForce)
            return RRNeedChars;
        }
        if ((ch == 'e' || ch == 'E') && _digits == Dec) {
          _litState = Exponent;
          advanceChar(buffer,position,loc);
          continue;
        }
        if ((ch == 'p' || ch == 'P') && _digits == Hex) {
          _litState = Exponent;
          advanceChar(buffer,position,loc);
          continue;
        }
        FloatingLiteralToken* floatToken = new FloatingLiteralToken;
        if (_digits == Dec)
          floatToken->setType(FloatingLiteralToken::TDecimal);
        if (_digits == Bin)
          floatToken->setType(FloatingLiteralToken::TBit);
        if (_digits == Hex)
          floatToken->setType(FloatingLiteralToken::THexaDecimal);
        _token = Token(floatToken);
        _litState = FSuf;
        continue;
      }
      case Exponent: {
        if (ch == '+' || ch == '-') {
          _litState = ExpNext;
          advanceChar(buffer,position,loc);
          continue;
        }
        _litState = ExpNext;
        continue;
      }
      case ExpNext: {
        if (isdigit(ch)) {
          advanceChar(buffer,position,loc);
          continue;
        }
        FloatingLiteralToken* floatToken = new FloatingLiteralToken;
        if (_digits == Dec)
          floatToken->setType(FloatingLiteralToken::TDecimal);
        if (_digits == Bin)
          floatToken->setType(FloatingLiteralToken::TBit);
        if (_digits == Hex)
          floatToken->setType(FloatingLiteralToken::THexaDecimal);
        _token = Token(floatToken);
        _litState = FSuf;
        continue;
      }
      case ISuf: {
        IntegerLiteralToken& intToken =
          (IntegerLiteralToken &)_token.getSContent();
        if (ch == 'l' || ch == 'L') {
          advanceChar(buffer,position,loc);
          if (intToken.isInt()) {
            intToken.mergeExtension(IntegerLiteralToken::ELong);
            continue;
          }
          if (intToken.isLong()) {
            intToken.mergeExtension(IntegerLiteralToken::ELongLong);
            continue;
          }
          std::string constant;
          _currentToken >> constant;
          std::ostringstream msg;
          msg << "Unknown suffix for integer literal constant: "
              << constant << ".";
          _errorList.push_back(Error(loc,msg.str()));
          // release the token
          delete _token.extractContent();
          return RRFinished;
        }
        if (ch == 'u' || ch == 'U') {
          advanceChar(buffer,position,loc);
          if (intToken.isUnsigned()) {
            std::string constant;
            _currentToken >> constant;
            std::ostringstream msg;
            msg << "Unknown suffix for integer literal constant: "
                << constant << ".";
            _errorList.push_back(Error(loc,msg.str()));
            // release the token
            delete _token.extractContent();
            return RRFinished;
          }
          intToken.mergeExtension(IntegerLiteralToken::EUnsigned);
          continue;
        }
        _currentToken >> intToken.value();
        return RRHasToken;
      }
      case FSuf: {
        FloatingLiteralToken& floatToken =
          (FloatingLiteralToken &)_token.getSContent();
        if (ch == 'l' || ch == 'L') {
          advanceChar(buffer,position,loc);
          if (floatToken.isReal()) {
            floatToken.mergeSuffix(FloatingLiteralToken::FSLongDouble);
            continue;
          }
          std::string constant;
          _currentToken >> constant;
          std::ostringstream msg;
          msg << "Unknown suffix for Real literal constant: "
              << constant << ".";
          _errorList.push_back(Error(loc,msg.str()));
          // release the token
          delete _token.extractContent();
          return RRFinished;
        }
        if (ch == 'd' || ch == 'D') {
          advanceChar(buffer,position,loc);
          if (floatToken.isReal()) {
            floatToken.mergeSuffix(FloatingLiteralToken::FSDouble);
            continue;
          }
          std::string constant;
          _currentToken >> constant;
          std::ostringstream msg;
          msg << "Unknown suffix for Real literal constant: "
              << constant << ".";
          _errorList.push_back(Error(loc,msg.str()));
          // release the token
          delete _token.extractContent();
          return RRFinished;
        }
        if (ch == 'f' || ch == 'F') {
          advanceChar(buffer,position,loc);
          if (floatToken.isReal()) {
            floatToken.mergeSuffix(FloatingLiteralToken::FSFloat);
            continue;
          }
          std::string constant;
          _currentToken >> constant;
          std::ostringstream msg;
          msg << "Unknown suffix for Real literal constant: "
              << constant << ".";
          _errorList.push_back(Error(loc,msg.str()));
          // release the token
          delete _token.extractContent();
          return RRFinished;
        }
        _currentToken >> floatToken.value();
        return RRHasToken;
      }
    }
  }
  return RRNeedChars;
}

bool Lexer::isUtf8SymbolStart(char ch) {
  _AcceptedUtf8Symbols.clear();
  std::string symb;
  symb = utf8_convert(0x2200);
  if (symb[0] == ch) {
    DLexer::KeywordToken tok;
    tok.setType(DLexer::KeywordToken::TForall);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2203);
  if(symb[0] == ch) {
    DLexer::KeywordToken tok;
    tok.setType(DLexer::KeywordToken::TExists);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2261);
  if(symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TEqual);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2262);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TDifferent);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2264);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TLessOrEqual);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2265);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TGreaterOrEqual);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2212);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TMinus);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x21D2);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TImplies);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x21D4);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TEquiv);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2227);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TLogicalAnd);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2228);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TLogicalOr);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x00AC);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TNot);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x22BB);
  if (symb[0] == ch) {
    DLexer::OperatorPunctuatorToken tok;
    tok.setType(DLexer::OperatorPunctuatorToken::TLogicalXor);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2208);
  if (symb[0] == ch) {
    DLexer::KeywordToken tok;
    tok.setType(DLexer::KeywordToken::TSubset);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x1D539);
  if (symb[0] == ch) {
    DLexer::KeywordToken tok;
    tok.setType(DLexer::KeywordToken::TBoolean);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x2124);
  if (symb[0] == ch) {
    DLexer::KeywordToken tok;
    tok.setType(DLexer::KeywordToken::TInteger);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  symb = utf8_convert(0x211D);
  if (symb[0] == ch) {
    DLexer::KeywordToken tok;
    tok.setType(DLexer::KeywordToken::TReal);
    _AcceptedUtf8Symbols.push_front(
      std::pair<std::string,AbstractToken>(symb,tok));
  }
  return !_AcceptedUtf8Symbols.empty();
}

Lexer::ReadResult
Lexer::readUtf8Symbol(const std::string& buffer, size_t& position,
                      location loc, bool doesForce) {
  size_t i = 1;
  while(!_AcceptedUtf8Symbols.empty()) {
    if (position+i>=buffer.size()) break;
    if (_AcceptedUtf8Symbols.size() == 1) {
      std::string symb = _AcceptedUtf8Symbols.front().first;
      size_t length = symb.size();
      while(i<length) {
        if (position+i >= buffer.size() || symb[i] != buffer[position+i]) {
          //TODO: use buffer[position] to determine the expected length of
          // the character in buffer
          size_t last = position+i>=buffer.size()?buffer.size()-1:position+i;
          std::ostringstream msg;
          msg << "unknown character: "
              << buffer.substr(position, last)
              << ".";
          _errorList.push_back(Error(loc,msg.str()));
          position=last+1;
          return RRContinueLexing;
        }
        i++;
      }
      position+=i;
      loc->charnum2+=i;
      _token = _AcceptedUtf8Symbols.front().second;
      return RRHasToken;
    } else {
      Utf8SymbolSet::iterator it = _AcceptedUtf8Symbols.begin();
      Utf8SymbolSet::iterator end = _AcceptedUtf8Symbols.end();
      while(it != end) {
        if ((*it).first[i] != buffer[position+i]) {
          Utf8SymbolSet::iterator tmp = it;
          it++;
          _AcceptedUtf8Symbols.erase(tmp,it);
        } else it++;
      }
    }
    i++;
  }
  std::ostringstream msg;
  msg << "unknown character: "
      << buffer.substr(position, position+i-1)
      << ".";
  _errorList.push_back(Error(loc,msg.str()));
  position+=i;
  return RRContinueLexing;
}

Lexer::ReadResult
Lexer::readCharLiteral(const std::string& buffer,
                       size_t& position, location loc, bool doesForce) {
  DLexer::CharacterLiteralToken* tok = new CharacterLiteralToken;
  _token = Token(tok);
  tok->setType(_charLitKind);
  size_t length = buffer.length();
  if (buffer[position] == '\\') {
    position++;
    loc->charnum2++;
    if (position>=length) {
      if (doesForce) {
        _errorList.push_back(Error(loc, "unterminated escape sequence"));
        return RRFinished;
      }
      return RRNeedChars;
    }
    switch(buffer[position]) {
      case '\'': tok->set_value('\''); break;
      case '"': tok->set_value('"'); break;
      case '?': tok->set_value('?'); break;
      case '\\': tok->set_value('\\'); break;
      case 'a': tok->set_value('\a'); break;
      case 'b': tok->set_value('\b'); break;
      case 'f': tok->set_value('\f'); break;
      case 'n': tok->set_value('\n'); break;
      case 'r': tok->set_value('\r'); break;
      case 't': tok->set_value('\t'); break;
      case 'v': tok->set_value('\v'); break;
      case 'x': {
        unsigned long long res = 0;
        while(true) {
          position++;
          loc->charnum2++;
          if (position>=length) {
            if (doesForce) {
              _errorList.push_back(Error(loc,"unterminated char literal"));
              return RRFinished;
            }
            return RRNeedChars;
          }
          if ('0'<=buffer[position] && buffer[position]<= '9')
            res = 16*res + (buffer[position]-'0');
          else if ('a'<=buffer[position]&&buffer[position]<='f')
            res = 16*res + 10 + buffer[position]-'a';
          else if ('A'<=buffer[position]&&buffer[position]<='F')
            res = 16*res + 10 + buffer[position]-'A';
          else { loc->charnum2--; position--; break; }
        }
        if (_charLitKind == DLexer::CharacterLiteralToken::TWideChar) {
          if (res > 1ULL << (8*sizeof(wchar_t))) {
            _errorList.push_back(
              Error(loc,"escape sequence not in wide char range"));
            return RRFinished;
          }
          tok->wchar_value() = res;
        } else {
          if (res > UCHAR_MAX) {
            _errorList.push_back(
              Error(loc,"escape sequence not in char range"));
            return RRFinished;
          }
          tok->char_value() = res;
        }
        break;
      }
      default:
        if ('0'<= buffer[position] && buffer[position] <= '9') {
          unsigned res = 0;
          for (int i = 0; i<3; i++) {
            if(position>=length) {
              if(doesForce) {
                _errorList.push_back(
                  Error(loc,"unterminated char literal"));
                return RRFinished;
              }
              return RRNeedChars;
            }
            if(buffer[position] < '0' || buffer[position] > '7') break;
            res = res*8 + (buffer[position]-'0');
            position++;
            loc->charnum2++;
          }
          position--;
          loc->charnum2--;
          if (_charLitKind == DLexer::CharacterLiteralToken::TWideChar) {
            // assume wchar_t is wide enough to accomodate the 0-512 range.
            tok->wchar_value() = res;
          } else {
            if (res > UCHAR_MAX) {
              _errorList.push_back(
                Error(loc,"escape sequence not in char range"));
              return RRFinished;
            }
            tok->char_value() = res;
          }
        } else
          // Just ignore the '\' (\ followed by an unknown character can be
          // supported in an implementation-defined manner as per [lex.ccon.3]
          tok->set_value(buffer[position]);
        break;
    }
  } else { tok->set_value(buffer[position]); }
  bool first_char = true;
  while(true) {
    position++;
    loc->charnum2++;
    if (position >= length) {
      if(doesForce) {
        _errorList.push_back(Error(loc,"unterminated char literal"));
        return RRFinished;
      }
      return RRNeedChars;
    }
    if (buffer[position] == '\'') {
      position++;
      _secondaryState = NULL;
      return RRHasToken;
    }
    if(first_char) {
      std::cerr <<
        loc->filename1 << ":" << loc->linenum1 << ":" << 
        loc->charnum1 << "-" << loc->charnum2 <<
        " Character literal has more than one char. Ignoring the other ones\n";
      first_char = false;
    }
    if (buffer[position] == '\\') position++;
  }
  assert(false);
}

Lexer::ReadResult
Lexer::readMain(const std::string& buffer, size_t& position, location loc,
    bool doesForce) {
  char ch = buffer[position];
  loc->linenum1 = loc->linenum2;
  loc->charnum1 = loc->charnum2+1;
  if (_hasNewlineToken) {
    _hasNewlineToken = false;
    ++loc->linenum1;
    loc->charnum1 = 1;
  };
  while (ch != '\0' && (isspace(ch) || ch == '@')) {
    ++position;
    if (ch == '\n') {
      ++loc->linenum1;
      loc->charnum1 = 1;
    }
    else
      ++loc->charnum1;
    ch = buffer[position];
  };
  if (ch == '\0')
    return doesForce ? RRFinished : RRNeedChars;

  loc->linenum2 = loc->linenum1;
  loc->charnum2 = loc->charnum1-1;
  switch (ch) {
    case '\\':
      _context = ch;
      _secondaryState = &Lexer::readProtectedToken;
      ++position;
      ++loc->charnum2;
      return (this->*_secondaryState)(buffer, position, loc, doesForce);
    case '<':
      if (!doesForce && buffer.length() - position < 4)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '=') {
        if (buffer[position+2] == '=') {
          if (buffer[position+3] == '>') {
            position += 4;
            loc->charnum2 += 4;
            _token = OperatorPunctuatorToken().setType(
                OperatorPunctuatorToken::TEquiv);
            return RRHasToken;
          };
        };
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TLessOrEqual);
        return RRHasToken;
      }
      else if (buffer[position+1] == '-') {
        if (buffer[position+2] == '-') {
          if (buffer[position+3] == '>') {
            position += 4;
            loc->charnum2 += 4;
            _token = OperatorPunctuatorToken().setType(
                OperatorPunctuatorToken::TBitEquiv);
            return RRHasToken;
          };
        };
      }
      else if (buffer[position+1] == '<') {
        if (buffer[position+2] == '=') {
          position += 3;
          loc->charnum2 += 3;
          _token = OperatorPunctuatorToken().setType(
              OperatorPunctuatorToken::TLeftShiftAssign);
          return RRHasToken;
        };
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TLeftShift);
        return RRHasToken;
      }
      else if (buffer[position+1] == ':') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TLTColon);
        return RRHasToken;
      };
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TLess);
      return RRHasToken;
    case '>':
      if (!doesForce && buffer.length() - position < 3)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TGreaterOrEqual);
        return RRHasToken;
      }
      else if (buffer[position+1] == '>') {
        if (buffer[position+2] == '=') {
          position += 3;
          loc->charnum2 += 3;
          _token = OperatorPunctuatorToken().setType(
              OperatorPunctuatorToken::TRightShiftAssign);
          return RRHasToken;
        };
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TRightShift);
        return RRHasToken;
      };
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TGreater);
      return RRHasToken;
    case '!':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TDifferent);
        return RRHasToken;
      }
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(OperatorPunctuatorToken::TNot);
      return RRHasToken;
    case '=':
      if (!doesForce && buffer.length() - position < 3)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '=') {
        if (buffer[position+2] == '>') {
          position += 3;
          loc->charnum2 += 3;
          _token = OperatorPunctuatorToken().setType(
              OperatorPunctuatorToken::TImplies);
          return RRHasToken;
        };
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TEqual);
        return RRHasToken;
      };
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TAssign);
      return RRHasToken;
    case '&':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '&') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TLogicalAnd);
        return RRHasToken;
      }
      else if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TBitAndAssign);
        return RRHasToken;
      }
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TAmpersand);
      return RRHasToken;
    case '|':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '|') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TLogicalOr);
        return RRHasToken;
      }
      else if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TBitOrAssign);
        return RRHasToken;
      }
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TBitOr);
      return RRHasToken;
    case '^':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '^') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TLogicalXor);
        return RRHasToken;
      }
      else if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TBitXorAssign);
        return RRHasToken;
      }
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TBitXor);
      return RRHasToken;
    case '-':
      if (!doesForce && buffer.length() - position < 3)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '-') {
        if (buffer[position+2] == '>') {
          position += 3;
          loc->charnum2 += 3;
          _token = OperatorPunctuatorToken().setType(
              OperatorPunctuatorToken::TBitImplies);
          return RRHasToken;
        };
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TMinusMinus);
        return RRHasToken;
      }
      else if (buffer[position+1] == '>') {
        if (buffer[position+2] == '*') {
          position += 3;
          loc->charnum2 += 3;
          _token = OperatorPunctuatorToken().setType(
              OperatorPunctuatorToken::TArrowStar);
          return RRHasToken;
        };
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TArrow);
        return RRHasToken;
      }
      else if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TMinusAssign);
        return RRHasToken;
      };
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TMinus);
      return RRHasToken;
    case '+':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '+') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TPlusPlus);
        return RRHasToken;
      }
      else if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TPlusAssign);
        return RRHasToken;
      };
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TPlus);
      return RRHasToken;
    case '{':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TOpenBrace);
      return RRHasToken;
    case '}':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TCloseBrace);
      return RRHasToken;
    case '[':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TOpenBracket);
      return RRHasToken;
    case ']':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TCloseBracket);
      return RRHasToken;
    case '(':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TOpenParen);
      return RRHasToken;
    case ')':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TCloseParen);
      return RRHasToken;
    case ';':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TSemiColon);
      return RRHasToken;
    case ':':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == ':') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TColonColon);
        return RRHasToken;
      }
      else if (buffer[position+1] == '>') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TColonGT);
        return RRHasToken;
      };
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TColon);
      return RRHasToken;
    case '.':
      if (!doesForce && buffer.length() - position < 3)
        return Lexer::RRNeedChars;
      if (buffer[position+1] == '.') {
        if (buffer[position+2] == '.') {
          position += 3;
          loc->charnum2 += 3;
          _token = OperatorPunctuatorToken().setType(
              OperatorPunctuatorToken::TEllipsis);
          return RRHasToken;
        };
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TRange);
        return RRHasToken;
      }
      else if (buffer[position+1] == '*') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TPunctStar);
        return RRHasToken;
      }
      else if (isdigit(buffer[position+1])) {
        advanceChar(buffer,position,loc);
        _secondaryState = &Lexer::readNumberToken;
        _digits = Dec;
        _litState = FracPart;
        return (this->*_secondaryState)(buffer, position, loc, doesForce);
      }
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TPunct);
      return RRHasToken;
    case '?':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TQuery);
      return RRHasToken;
    case '*':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      else if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TMultAssign);
        return RRHasToken;
      }
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TStar);
      return RRHasToken;
    case '/':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      else if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TDivideAssign);
        return RRHasToken;
      }
      else if (buffer[position+1] == '/') {
        position += 2;
        loc->charnum2 += 2;
        _context = '/';
        _secondaryState = &Lexer::readEndComment; 
        return (this->*_secondaryState)(buffer, position, loc, doesForce);
      }
      else if (buffer[position+1] == '*') {
        position += 2;
        loc->charnum2 += 2;
        _context = '*';
        _secondaryState = &Lexer::readEndComment; 
        return (this->*_secondaryState)(buffer, position, loc, doesForce);
      };
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TDivide);
      return RRHasToken;
    case '%':
      if (!doesForce && buffer.length() - position < 2)
        return Lexer::RRNeedChars;
      else if (buffer[position+1] == '=') {
        position += 2;
        loc->charnum2 += 2;
        _token = OperatorPunctuatorToken().setType(
            OperatorPunctuatorToken::TModuloAssign);
        return RRHasToken;
      }
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TProcent);
      return RRHasToken;
    case '~':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TTilda);
      return RRHasToken;
    case ',':
      position += 1;
      loc->charnum2 += 1;
      _token = OperatorPunctuatorToken().setType(
          OperatorPunctuatorToken::TComma);
      return RRHasToken;
    case '\'':
      position+=1;
      loc->charnum2+=1;
      _secondaryState = &Lexer::readCharLiteral;
      _charLitKind = DLexer::CharacterLiteralToken::TChar;
      return(this->*_secondaryState)(buffer,position,loc,doesForce);
    case 'L':
      if (buffer[position+1] == '\'') {
        position+=2;
        loc->charnum2+=2;
        _secondaryState = &Lexer::readCharLiteral;
        _charLitKind = DLexer::CharacterLiteralToken::TWideChar;
        return(this->*_secondaryState)(buffer,position,loc,doesForce);
      } else break;
    case '0':
      advanceChar(buffer,position,loc);
      _secondaryState = &Lexer::readNumberToken;
      _litState = LeadZero;
      return (this->*_secondaryState)(buffer,position,loc,doesForce);
  };
  if (isalpha(ch) || ch == '_') {
    _secondaryState = &Lexer::readIdentifierToken; 
    return (this->*_secondaryState)(buffer, position, loc, doesForce);
  };
  if (isdigit(ch)) {
    advanceChar(buffer,position,loc);
    _secondaryState = &Lexer::readNumberToken;
    _digits = Dec;
    _litState = Digits;
    return (this->*_secondaryState)(buffer, position, loc, doesForce);
  };

  if (isUtf8SymbolStart(ch)) {
    _secondaryState = &Lexer::readUtf8Symbol;
    return (this->*_secondaryState)(buffer,position,loc,doesForce);
  }
  // illegal character. Just add an error message.
  ++position; // otherwise, we're stuck in an infinite loop
  std::ostringstream msg;
  msg << "unknown character: " << ch << ".";
  _errorList.push_back(Error(loc, msg.str()));
  return RRContinueLexing;
}

} // end of namespace Acsl

